/*
    Gaudium  - an HTTPSERVER for Nodejs
    Copyright (C)  2011  freefeathers | Tierracycle Cooperative

    Contributors:
                   Nathaniel Jayme <nbjayme@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


var _libhttp = require('http')
var _libsys = require('sys')
var _libfs = require('fs')
var _libpath = require('path')
var _liburl = require('url')
var _libMime = require('./mime.js')
try
{
   var _libuuid = require('node-uuid')
}
catch(e)
{
   var _libuuid = require('../node-uuid')
}

var Worker = {}

//sample config
// please copy the example provided in the "example" folder

var vhosts={}

vhosts["default"]={
   docRoot: __dirname
}

/*
vhosts["www.example.com"]={
   "docRoot": __dirname
}
*/

/* singleton */
Worker.Config={
     port:8081,
     docIndex:["index.html" , "index.htm", "index.nodejs"] ,
     nodeScript:"nodejs",
     hosts: vhosts,
     errDocs: "err",
     defaultMime:"text/plain"
}

/* singleton */
Worker.State = {
     _isRunning:false,

    isRunning:function(flag)
    {
         if ( flag != null )
         {
               this._isRunning = flag
         }
         return this._isRunning
    }
}

/* singleton */
Worker.Util ={

    clearRequireCache: function (libName)
    {
          var _lib = libName.match(/[^/]+$/)
          if ( _lib == null )
          {
              return
          }

          _lib = _lib[0]

          var r = new RegExp( '/' + _lib + '$')
          for (var s in require.cache)
          {
               if ( r.exec(s) )
               {
                   delete require.cache[s]
               }
          }
    },

    getErrFileContent:function(basePath,code)
    {
          var errFile = basePath + "/" + Worker.Config.errDocs + "/doc" + code + ".html"
          return this.getFileContent(errFile)
    },

    getFileContent:function( fPath )
    {
          var data = null
          try
          {
              var stats = _libfs.statSync(fPath)
              if (stats.isFile() )
              {
                  data = _libfs.readFileSync(fPath,"UTF-8")
              }
          }
          catch (e){  }
          return data
    }
}


Worker.Message=function( connection ){
     // private variables
     var _self = this
     var _connection = connection

     // public methods and properties

     this.freeMem = function()
     {
         _connection = null
         _self = null
     }

     this.getSelf = function()
     {
          return _self
     }
     this.getConnection=function()
     {
         return _connection
     }
     this._header = {}
}

Worker.Message.prototype = {
    appendHeader:function(collection)
    {
           for ( item in collection )
           {
              this._header[item] = collection[item]
           }
    },

    clearHeader:function()
    {
           for (item in this._header)
           {
               delete this._header[ item ]
           }
    },

    getHeader:function()
    {
          return this._header
    },

    getResponse:function()
    {
          return this.getConnection().getResponse()
    },

    initHeader:function( content_type )
    {
       if ( content_type == null )
       {
              content_type = Worker.Config.defaultMime
        }
        this.clearHeader()
        this.appendHeader( { 'content-type' : content_type } )
        this.appendHeader( { 'transfer-encoding' : 'chunked' } )
    },

    send:function(code,data)
    {
        var resp = this.getResponse()
        resp.writeHead( code, this.getHeader() )
        resp.end( data)
        this.getConnection().setStatus('done')
    },

    sendOk:function( data, content_type)
    {
          this.initHeader( content_type )
          this.send( 200,data)
    },

    sendFileContent:function(fPath, content_type )
    {
           var _self = this.getSelf()
           var extension = _libpath.extname(fPath)
           if (extension) extension = extension.substr(1)
           if (extension == Worker.Config.nodeScript )
           {
                   setTimeout( function() {  Worker.Events.ExecuteScript(_self,fPath)  }, 150 )
                   return
           }

           if (extension in _libMime.Mime )
           {
                var mime = _libMime.Mime[ extension ]
           }

           if (content_type == null )
           {
               content_type = mime
           }

           var resp = this.getConnection().getResponse()
           this.initHeader(content_type)
           resp.writeHead( 200, this.getHeader() )
           
           if (content_type.match(/^text/) == null )
           {
              var data = _libfs.readFileSync(fPath)
              resp.end( data , 'binary') 
           }
           else
           {
              var data = _libfs.readFileSync(fPath,'UTF-8')
              resp.end( data ) 
           }
           this.getConnection().setStatus('done')
    },

    sendUnhandledException:function(e)
    {
         this.initHeader("text/html")
         var data = e.stack.replace('\n', '<br />')
         var dataHtml = Worker.Util.getErrFileContent(this.getConnection().getDocRoot(),"UnhandledException")
         if ( dataHtml )
         {
               data = dataHtml +  "<br /> <br />Stack trace resut: <br /> <br />" + data
         }

         this.send(200,data)
    },

    sendNotFound:function()
    {
          this.initHeader( "text/html ")
          var data = Worker.Util.getErrFileContent(this.getConnection().getDocRoot(),404)
          if (data == null )
          {
              data = "Resource not found" 
          }
          this.send( 404, data)
    }

}

Worker.Connection =function(req,resp){
    // protected variables
     var _id =   _libuuid()
     var _self = this
     var _message = new Worker.Message( _self )

     // public methods and properties
     this.freeMem=function()
     {
           _self = null
           _message.freeMem()
           delete _message
     }

     this.getMessageService = function()
     {
        return _message
     }

     this.getId = function()
     {
          return _id
     }
     this.getSelf = function()
     {
          return _self
     }

     this._status = "onhold"
     this._request=req
     this._response=resp

     // to be assigned when parsing the URI base on Worker.Config.hosts entry
     this._host=null
     this._port=null
     this._requestURI=null
     this._docRoot=""

     this.initialize(req,resp) 
}

Worker.Connection.prototype={

    initialize:function(req,resp){
        this._request = req
        this._resp = resp

        var parsed = req.headers['host'].split(':')
        this._host = parsed[0]
        this._port = parsed[1]

        this._requestURI = _liburl.parse( this._request.url, true, true )

        // find docroot
        this._docRoot = null

        if (this._host == '127.0.0.1' )
        {
              this._host = 'localhost'
        }

        if ( this._host == 'localhost' )
        {
             this._docRoot  = Worker.Config.hosts[ "default" ].docRoot
        }

        if ( this._host in Worker.Config.hosts )
        {
             this._docRoot = Worker.Config.hosts[ this._host ].docRoot
        }

        this.Message= new Worker.Message(this)

        if ( this._docRoot == null )
        {
            this.Message.sendNotFound()
        }
    },

    getHost:function()
    {
          return this._host
    },

    getPort:function()
    {
          return this._port
    },

    getDocRoot:function()
    {
          return this._docRoot
    },

    setResponse:function(resp){
        this._response = resp
    },

    getRequest:function(){
        return this._request
    },

    getResponse:function(){
        return this._response
    },

    getRequestURI:function(){
         return this._requestURI
    },

    getStatus:function(){
        return this._status
    },

    setStatus:function( status )
    {
        this._status = status
    },

    isPathSecure:function(fullPath){
        var requestPath = _libpath.normalize( fullPath )
        var re = new RegExp( "^" + this.getDocRoot() )
        if (  re.exec( requestPath ) == null ) {
             return false
        }
        return true
    },
    
    Execute:function(){
         this.setStatus( 'executing' )

         var requestURI  = this.getRequestURI()

         var fullPath = _libpath.normalize ( this.getDocRoot() + requestURI.pathname )

         if ( ! this.isPathSecure (fullPath ) )
         {
               this.getMessageService().sendNotFound()
         }

          try {
               var stats = _libfs.statSync(fullPath)
          }
          catch(e)
          {
               this.getMessageService().sendNotFound()
               return
          }

          if ( stats.isDirectory() )
          {
              var docIndeces = Worker.Config.docIndex

              for (var nIndex = 0, maxIndex= docIndeces.length; nIndex < maxIndex; nIndex++ )
              {
                  var indexFile = fullPath + docIndeces[nIndex]
                  try
                  {
                     stats = _libfs.statSync(indexFile)
                     fullPath = indexFile
                     break
                  }
                  catch(e)
                  {
                      continue
                  }
              }

              if ( ! stats.isFile() )
              {
                  this.getMessageService().sendNotFound()
                 return
              }
          }

          if (stats.isFile() )
          {
              this.getMessageService().sendFileContent(fullPath)
              return
          }

          this.getMessageService().sendNotFound()
    }
}


/* singleton */
Worker.Manager=
{
     _workers:{},  // client connections
     _activeCount:0,
     _cleanerTick:1200,

     CreateAWorker:function(req,resp)
     {
          setTimeout( function() { Worker.Manager.CleanDoneWorkers() } ,this._cleanerTick )
          this._activeCount++
          var connection = new Worker.Connection(req,resp)
          var id = connection.getId()

          this._workers[ id ] = connection
          return this._workers[ id ]
     },

     CleanAWorker:function( id )
     {
          var connection = this.getAWorker( id )
          if (connection == null )
          {
             return
          }
          this._workers[ id ].freeMem()
          delete this._workers[ id ]
          this._activeCount--
     },

     getAWorker:function( id )
     {
           if ( id in this._workers )
           {
                return this._workers[ id ]
           }
           return null
     },

     CleanDoneWorkers:function(id)
     {
           for (var id in this._workers )
           {
                oWorker = this._workers[ id ]
                if ( oWorker.getStatus() != 'done' ) 
                {
                   continue
                }
                this.CleanAWorker( id )
           }
           if (this._activeCount > 0 )
           {
              setTimeout( function() { Worker.Manager.CleanDoneWorkers() } , this._cleanerTick)
           }
     }

}

/* singleton */
Worker.Events={
    Request:function(req,resp)
    {
       var workerConnection =  Worker.Manager.CreateAWorker( req, resp )
       workerConnection.Execute()
    },

    Activate:function(config)
    {
           if(   Worker.State.isRunning() )
           {
                 return
           }

           if ( config != null )
           {
                for (item in config)
                {
                     Worker.Config[item]  = config[item]
                }
           }

           Worker.State.isRunning( true )
           console.log("Starting Web server at port " + Worker.Config.port )
           console.log("Hosted domains: " )
           for ( var h  in Worker.Config.hosts )
           {
                console.log ( ' domain ( ' + h + ')  -----> ' + Worker.Config.hosts[ h ].docRoot )
           }

           Worker.Service = new _libhttp.Server() 
           Worker.Service.listen(Worker.Config.port)
           Worker.Service.on("request", function(req,resp) {
                       setTimeout( function() { Worker.Events.Request(req,resp) }, 150) 
            })
    },

    ExecuteScript:function( wwwService, fPath )
    {
        try
        {
           var nodeScript = require( fPath )
           Worker.Util.clearRequireCache( fPath )
           nodeScript.Execute( wwwService )
        }
        catch(e)
        {
            this.sendUnhandledException( e )
            Worker.Util.clearRequireCache( fPath )
        }
    },

}

exports.GaudiumServer=function()
{
    obj = {
         isRunning:function()
         {
             return Worker.State.isRunning()
         },

         startService:function(config)
         {
             Worker.Events.Activate(config)
         }
   }
   return obj
}

